/**
 * Parametric Modeling (PMG) extension for SketchUp.
 *
 * @copyright © 2021 Samuel Tallet
 *
 * @licence GNU General Public License 3.0
 */

PMG = {}

PMG.NodesEditor = {}

PMG.NodesEditor.initializeSockets = () => {

    PMG.NodesEditor.sockets = {

        number: new Rete.Socket('Number'),
        groups: new Rete.Socket('Group(s)'),
        point: new Rete.Socket('Point'),
        vector: new Rete.Socket('Vector')

    }

}

PMG.NodesEditor.initializeSockets()

PMG.cloneObject = (object) => {
    return JSON.parse(JSON.stringify(object))
}

PMG.NodesEditor.initializeControls = () => {

    PMG.NodesEditor.controls = {}

    PMG.NodesEditor.controls['number'] = {

        props: ['emitter', 'ikey', 'getData', 'putData', 'placeholder', 'readonly'],

        template: '<input type="number" :placeholder="placeholder" :readonly="readonly" :value="value" @input="change($event)" @pointerdown.stop="" @pointermove.stop="" />',
        
        data() {

            return {
                value: 0
            }

        },

        methods: {

            change(event) {

                this.value = + event.target.value
                this.update()

            },

            update() {

                if ( this.ikey )
                    this.putData(this.ikey, this.value)

                this.emitter.trigger('process')

            }

        },

        mounted() {
            this.value = this.getData(this.ikey)
        }
        
    }

    PMG.NodesEditor.controls['text'] = {

        props: ['emitter', 'ikey', 'getData', 'putData', 'placeholder', 'readonly'],

        template: '<input type="text" :placeholder="placeholder" :readonly="readonly" :value="value" @input="change($event)" @pointerdown.stop="" @pointermove.stop="" />',
        
        data() {

            return {
                value: ''
            }

        },

        methods: {

            change(event) {

                this.value = event.target.value
                this.update()

            },

            update() {

                if ( this.ikey )
                    this.putData(this.ikey, this.value)

                this.emitter.trigger('process')

            }

        },

        mounted() {
            this.value = this.getData(this.ikey)
        }
        
    }

    PMG.NodesEditor.controls['checkbox'] = {

        props: ['emitter', 'ikey', 'getData', 'putData', 'label'],

        template:
        `<div class="checkbox-control">
            <input type="checkbox" @change="change($event)" :checked="checked" />
            {{ label }}
        </div>`,

        data() {

            return {
                checked: this.getData(this.ikey),
                label: this.label
            }
            
        },

        methods: {

            change(event) {

                this.checked = event.target.checked
                this.update()

            },

            update() {

                if ( this.ikey )
                    this.putData(this.ikey, this.checked)

                this.emitter.trigger('process')

            }

        },

        mounted() {
            this.checked = this.getData(this.ikey)
        }
        
    }

    PMG.NodesEditor.controls['material'] = {

        props: ['emitter', 'ikey', 'getData', 'putData'],

        template:
        `<select @change="change($event)">
            <option value="">材质...</option>
            <option v-for="material in materials" v-bind:value="material.name" v-bind:selected="material.selected">
                {{ material.display_name }}
            </option>
        </select>`,

        data() {

            var nodeSelectedMaterial = this.getData(this.ikey)
            var nodeSelectableMaterials = PMG.cloneObject(SketchUpMaterials)

            nodeSelectableMaterials.forEach(nodeSelectableMaterial => {

                if ( nodeSelectableMaterial.name == nodeSelectedMaterial ) {
                    nodeSelectableMaterial.selected = 'selected'
                } else {
                    nodeSelectableMaterial.selected = ''
                }

            })

            return {
                materials: nodeSelectableMaterials
            }
            
        },

        methods: {

            change(event) {

                this.value = event.target.value
                this.update()

            },

            update() {

                if ( this.ikey )
                    this.putData(this.ikey, this.value)

                this.emitter.trigger('process')

            }

        },

        mounted() {
            this.value = this.getData(this.ikey)
        }
        
    }

    PMG.NodesEditor.controls['layer'] = {

        props: ['emitter', 'ikey', 'getData', 'putData'],

        template:
        `<select @change="change($event)">
            <option value="">图层/标签...</option>
            <option v-for="layer in layers" v-bind:value="layer.name" v-bind:selected="layer.selected">
                {{ layer.display_name }}
            </option>
        </select>`,

        data() {

            var nodeSelectedLayer = this.getData(this.ikey)
            var nodeSelectableLayers = PMG.cloneObject(SketchUpLayers)

            nodeSelectableLayers.forEach(nodeSelectableLayer => {

                if ( nodeSelectableLayer.name == nodeSelectedLayer ) {
                    nodeSelectableLayer.selected = 'selected'
                } else {
                    nodeSelectableLayer.selected = ''
                }

            })

            return {
                layers: nodeSelectableLayers
            }
            
        },

        methods: {

            change(event) {

                this.value = event.target.value
                this.update()

            },

            update() {

                if ( this.ikey )
                    this.putData(this.ikey, this.value)

                this.emitter.trigger('process')

            }

        },

        mounted() {
            this.value = this.getData(this.ikey)
        }
        
    }

}

PMG.NodesEditor.initializeControls()

class NumberReteControl extends Rete.Control {

    constructor(emitter, ikey, placeholder, readonly) {

        super(ikey)
        this.component = PMG.NodesEditor.controls.number
        this.props = { emitter, ikey, placeholder, readonly }

    }
  
    setValue(value) {
        this.vueContext.value = value
    }

}

class TextReteControl extends Rete.Control {

    constructor(emitter, ikey, placeholder, readonly) {

        super(ikey)
        this.component = PMG.NodesEditor.controls.text
        this.props = { emitter, ikey, placeholder, readonly }

    }
  
    setValue(value) {
        this.vueContext.value = value
    }

}

class CheckBoxReteControl extends Rete.Control {

    constructor(emitter, ikey, label) {

        super(ikey)
        this.component = PMG.NodesEditor.controls.checkbox
        this.props = { emitter, ikey, label }

    }
  
    setValue(checked) {
        this.vueContext.checked = checked
    }

}

class MaterialReteControl extends Rete.Control {

    constructor(emitter, key, _readonly) {

        super(key)
        this.component = PMG.NodesEditor.controls.material
        this.props = { emitter, ikey: key }

    }
  
    setValue(value) {
        this.vueContext.value = value
    }

}

class LayerReteControl extends Rete.Control {

    constructor(emitter, key, _readonly) {

        super(key)
        this.component = PMG.NodesEditor.controls.layer
        this.props = { emitter, ikey: key }

    }
  
    setValue(value) {
        this.vueContext.value = value
    }

}

class DrawBoxReteComponent extends Rete.Component {

    constructor(){
        super('绘制盒子')
    }

    builder (node) {

        var width = new Rete.Input('width', '宽度', PMG.NodesEditor.sockets.number)
        var depth = new Rete.Input('depth', '深度', PMG.NodesEditor.sockets.number)
        var height = new Rete.Input('height', '高度', PMG.NodesEditor.sockets.number)

        var group = new Rete.Output('groups', '群组', PMG.NodesEditor.sockets.groups)

        return node
            .addInput(width)
            .addInput(depth)
            .addInput(height)
            .addControl(new TextReteControl(this.editor, 'name', '名称'))
            .addControl(new MaterialReteControl(this.editor, 'material'))
            .addControl(new LayerReteControl(this.editor, 'layer'))
            .addOutput(group)

    }

    worker(_node, _inputs, outputs) {
        outputs['groups'] = ''
    }

}

class DrawPrismReteComponent extends Rete.Component {

    constructor(){
        super('绘制棱柱')
    }

    builder (node) {

        var radius = new Rete.Input('radius', '半径 ', PMG.NodesEditor.sockets.number)
        var height = new Rete.Input('height', '高度', PMG.NodesEditor.sockets.number)
        var sides = new Rete.Input('sides', '边数', PMG.NodesEditor.sockets.number)

        var group = new Rete.Output('groups', '群组', PMG.NodesEditor.sockets.groups)

        return node
            .addInput(radius)
            .addInput(height)
            .addInput(sides)
            .addControl(new TextReteControl(this.editor, 'name', '名称'))
            .addControl(new MaterialReteControl(this.editor, 'material'))
            .addControl(new LayerReteControl(this.editor, 'layer'))
            .addOutput(group)

    }

    worker(_node, _inputs, outputs) {
        outputs['groups'] = ''
    }

}

class DrawCylinderReteComponent extends Rete.Component {

    constructor(){
        super('绘制圆柱')
    }

    builder (node) {

        var radius = new Rete.Input('radius', '半径', PMG.NodesEditor.sockets.number)
        var height = new Rete.Input('height', '高度', PMG.NodesEditor.sockets.number)
        var segments = new Rete.Input('segments', '段数', PMG.NodesEditor.sockets.number)

        var group = new Rete.Output('groups', '群组', PMG.NodesEditor.sockets.groups)

        return node
            .addInput(radius)
            .addInput(height)
            .addInput(segments)
            .addControl(new TextReteControl(this.editor, 'name', '名称'))
            .addControl(new MaterialReteControl(this.editor, 'material'))
            .addControl(new LayerReteControl(this.editor, 'layer'))
            .addOutput(group)

    }

    worker(_node, _inputs, outputs) {
        outputs['groups'] = ''
    }

}

class DrawPyramidReteComponent extends Rete.Component {

    constructor(){
        super('绘制棱锥')
    }

    builder (node) {

        var radius = new Rete.Input('radius', '半径', PMG.NodesEditor.sockets.number)
        var height = new Rete.Input('height', '高度', PMG.NodesEditor.sockets.number)
        var sides = new Rete.Input('sides', '边数', PMG.NodesEditor.sockets.number)

        var group = new Rete.Output('groups', '群组', PMG.NodesEditor.sockets.groups)

        return node
            .addInput(radius)
            .addInput(height)
            .addInput(sides)
            .addControl(new TextReteControl(this.editor, 'name', '名称'))
            .addControl(new MaterialReteControl(this.editor, 'material'))
            .addControl(new LayerReteControl(this.editor, 'layer'))
            .addOutput(group)

    }

    worker(_node, _inputs, outputs) {
        outputs['groups'] = ''
    }

}

class DrawConeReteComponent extends Rete.Component {

    constructor(){
        super('绘制圆锥')
    }

    builder (node) {

        var radius = new Rete.Input('radius', '半径', PMG.NodesEditor.sockets.number)
        var height = new Rete.Input('height', '高度', PMG.NodesEditor.sockets.number)
        var segments = new Rete.Input('segments', '段数', PMG.NodesEditor.sockets.number)

        var group = new Rete.Output('groups', '群组', PMG.NodesEditor.sockets.groups)

        return node
            .addInput(radius)
            .addInput(height)
            .addInput(segments)
            .addControl(new TextReteControl(this.editor, 'name', '名称'))
            .addControl(new MaterialReteControl(this.editor, 'material'))
            .addControl(new LayerReteControl(this.editor, 'layer'))
            .addOutput(group)

    }

    worker(_node, _inputs, outputs) {
        outputs['groups'] = ''
    }

}

class DrawSphereReteComponent extends Rete.Component {

    constructor(){
        super('绘制球体')
    }

    builder (node) {

        var radius = new Rete.Input('radius', '半径', PMG.NodesEditor.sockets.number)

        var group = new Rete.Output('groups', '群组', PMG.NodesEditor.sockets.groups)

        return node
            .addInput(radius)
            .addControl(new TextReteControl(this.editor, 'name', '名称'))
            .addControl(new MaterialReteControl(this.editor, 'material'))
            .addControl(new LayerReteControl(this.editor, 'layer'))
            .addOutput(group)

    }

    worker(_node, _inputs, outputs) {
        outputs['groups'] = ''
    }

}

class NumberReteComponent extends Rete.Component {

    constructor() {
        super('数量')
    }

    builder(node) {

        var number = new Rete.Output('number', '数量', PMG.NodesEditor.sockets.number)
        return node.addControl(new NumberReteControl(this.editor, 'number')).addOutput(number)

    }

    worker(node, _inputs, outputs) {
        outputs['number'] = node.data.number
    }
}

class AddReteComponent extends Rete.Component {

    constructor(){
        super('加')
    }

    builder (node) {

        var number1 = new Rete.Input('number1', '数量', PMG.NodesEditor.sockets.number)
        var number2 = new Rete.Input('number2', '数量', PMG.NodesEditor.sockets.number)

        var number = new Rete.Output('number', '数量', PMG.NodesEditor.sockets.number)

        number1.addControl(new NumberReteControl(this.editor, 'number1'))
        number2.addControl(new NumberReteControl(this.editor, 'number2'))

        return node
            .addInput(number1)
            .addInput(number2)
            .addControl(new NumberReteControl(this.editor, 'preview', '', true))
            .addOutput(number)

    }

    worker(node, inputs, outputs) {

        var number1 = inputs['number1'].length ? inputs['number1'][0] : node.data.number1
        var number2 = inputs['number2'].length ? inputs['number2'][0] : node.data.number2
        var sum = number1 + number2
        
        this.editor.nodes.find(n => n.id == node.id).controls.get('preview').setValue(sum)
        outputs['number'] = sum

    }

}

class SubtractReteComponent extends Rete.Component {

    constructor(){
        super('减')
    }

    builder (node) {

        var number1 = new Rete.Input('number1', '数量', PMG.NodesEditor.sockets.number)
        var number2 = new Rete.Input('number2', '数量', PMG.NodesEditor.sockets.number)

        var number = new Rete.Output('number', '数量', PMG.NodesEditor.sockets.number)

        number1.addControl(new NumberReteControl(this.editor, 'number1'))
        number2.addControl(new NumberReteControl(this.editor, 'number2'))

        return node
            .addInput(number1)
            .addInput(number2)
            .addControl(new NumberReteControl(this.editor, 'preview', '', true))
            .addOutput(number)

    }

    worker(node, inputs, outputs) {

        var number1 = inputs['number1'].length ? inputs['number1'][0] : node.data.number1
        var number2 = inputs['number2'].length ? inputs['number2'][0] : node.data.number2
        var diff = number1 - number2
        
        this.editor.nodes.find(n => n.id == node.id).controls.get('preview').setValue(diff)
        outputs['number'] = diff

    }

}

class PointReteComponent extends Rete.Component {

    constructor() {
        super('指向')
    }

    builder(node) {

        var inputPointX = new Rete.Input('x', 'X轴', PMG.NodesEditor.sockets.number)
        inputPointX.addControl(new NumberReteControl(this.editor, 'x', 'X轴'))

        var inputPointY = new Rete.Input('y', 'Y轴', PMG.NodesEditor.sockets.number)
        inputPointY.addControl(new NumberReteControl(this.editor, 'y', 'Y轴'))

        var inputPointZ = new Rete.Input('z', 'Z轴', PMG.NodesEditor.sockets.number)
        inputPointZ.addControl(new NumberReteControl(this.editor, 'z', 'Z轴'))

        var outputPoint = new Rete.Output('point', '指向', PMG.NodesEditor.sockets.point)

        return node
            .addInput(inputPointX)
            .addInput(inputPointY)
            .addInput(inputPointZ)
            .addOutput(outputPoint)

    }

    worker(node, _inputs, outputs) {
        outputs['point'] = { x: node.data.x, y: node.data.y, z: node.data.z }
    }
}

class VectorReteComponent extends Rete.Component {

    constructor() {
        super('向量')
    }

    builder(node) {

        var inputVectorX = new Rete.Input('x', 'X轴', PMG.NodesEditor.sockets.number)
        inputVectorX.addControl(new NumberReteControl(this.editor, 'x', 'X轴'))

        var inputVectorY = new Rete.Input('y', 'Y轴', PMG.NodesEditor.sockets.number)
        inputVectorY.addControl(new NumberReteControl(this.editor, 'y', 'Y轴'))

        var inputVectorZ = new Rete.Input('z', 'Z轴', PMG.NodesEditor.sockets.number)
        inputVectorZ.addControl(new NumberReteControl(this.editor, 'z', 'Z轴'))

        var outputVector = new Rete.Output('vector', '向量', PMG.NodesEditor.sockets.vector)

        return node
            .addInput(inputVectorX)
            .addInput(inputVectorY)
            .addInput(inputVectorZ)
            .addOutput(outputVector)

    }

    worker(node, _inputs, outputs) {
        outputs['vector'] = { x: node.data.x, y: node.data.y, z: node.data.z }
    }
}

class PushPullReteComponent extends Rete.Component {

    constructor(){
        super('推拉')
    }

    builder (node) {

        var inputGroups = new Rete.Input('groups', '群组', PMG.NodesEditor.sockets.groups)
        var inputDistance = new Rete.Input('distance', '距离', PMG.NodesEditor.sockets.number)
        inputDistance.addControl(new NumberReteControl(this.editor, 'distance', '距离'))
        var inputDirection = new Rete.Input('direction', '方向', PMG.NodesEditor.sockets.vector)

        var outputGroups = new Rete.Output('groups', '群组', PMG.NodesEditor.sockets.groups)

        return node
            .addInput(inputGroups)
            .addInput(inputDistance)
            .addControl(new CheckBoxReteControl(this.editor, 'increment_distance', '增量距离'))
            .addInput(inputDirection)
            .addOutput(outputGroups)

    }

    worker(_node, _inputs, outputs) {
        outputs['groups'] = ''
    }

}

class MoveReteComponent extends Rete.Component {

    constructor(){
        super('移动')
    }

    builder (node) {

        var inputGroups = new Rete.Input('groups', '群组', PMG.NodesEditor.sockets.groups)
        var inputPoint = new Rete.Input('point', '指向', PMG.NodesEditor.sockets.point)

        var outputGroups = new Rete.Output('groups', '群组', PMG.NodesEditor.sockets.groups)

        return node
            .addInput(inputGroups)
            .addInput(inputPoint)
            .addOutput(outputGroups)

    }

    worker(_node, _inputs, outputs) {
        outputs['groups'] = ''
    }

}

class RotateReteComponent extends Rete.Component {

    constructor(){
        super('旋转')
    }

    builder (node) {

        var inputGroups = new Rete.Input('groups', '群组', PMG.NodesEditor.sockets.groups)
        var inputCenter = new Rete.Input('center', '中心', PMG.NodesEditor.sockets.point)
        var inputAxis = new Rete.Input('axis', '轴', PMG.NodesEditor.sockets.vector)
        var inputAngle = new Rete.Input('angle', '角度', PMG.NodesEditor.sockets.number)
        inputAngle.addControl(new NumberReteControl(this.editor, 'angle', '角度'))

        var outputGroups = new Rete.Output('groups', '群组', PMG.NodesEditor.sockets.groups)

        return node
            .addInput(inputGroups)
            .addInput(inputCenter)
            .addInput(inputAxis)
            .addInput(inputAngle)
            .addOutput(outputGroups)

    }

    worker(_node, _inputs, outputs) {
        outputs['groups'] = ''
    }

}

class ScaleReteComponent extends Rete.Component {

    constructor(){
        super('比例')
    }

    builder (node) {

        var inputGroups = new Rete.Input('groups', '群组', PMG.NodesEditor.sockets.groups)
        var inputPoint = new Rete.Input('point', '指向', PMG.NodesEditor.sockets.point)
        var inputXFactor = new Rete.Input('x_factor', 'X 系数', PMG.NodesEditor.sockets.number)
        inputXFactor.addControl(new NumberReteControl(this.editor, 'x_factor', 'X 系数'))
        var inputYFactor = new Rete.Input('y_factor', 'Y 系数', PMG.NodesEditor.sockets.number)
        inputYFactor.addControl(new NumberReteControl(this.editor, 'y_factor', 'Y 系数'))
        var inputZFactor = new Rete.Input('z_factor', 'Z 系数', PMG.NodesEditor.sockets.number)
        inputZFactor.addControl(new NumberReteControl(this.editor, 'z_factor', 'Z 系数'))

        var outputGroups = new Rete.Output('groups', '群组', PMG.NodesEditor.sockets.groups)

        return node
            .addInput(inputGroups)
            .addInput(inputPoint)
            .addInput(inputXFactor)
            .addInput(inputYFactor)
            .addInput(inputZFactor)
            .addOutput(outputGroups)

    }

    worker(_node, _inputs, outputs) {
        outputs['groups'] = ''
    }

}

class CopyReteComponent extends Rete.Component {

    constructor(){
        super('复制')
    }

    builder (node) {

        var inputGroups = new Rete.Input('groups', '群组', PMG.NodesEditor.sockets.groups)
        var inputCopies = new Rete.Input('copies', '副本', PMG.NodesEditor.sockets.number)
        inputCopies.addControl(new NumberReteControl(this.editor, 'copies', '副本'))

        var outputGroups = new Rete.Output('groups', '群组', PMG.NodesEditor.sockets.groups)

        return node
            .addInput(inputGroups)
            .addInput(inputCopies)
            .addControl(new CheckBoxReteControl(this.editor, 'output_original', '输出原始'))
            .addOutput(outputGroups)

    }

    worker(_node, _inputs, outputs) {
        outputs['groups'] = ''
    }

}

PMG.codeName = 'ParametricModeling'

PMG.NodesEditor.initializeEditor = () => {

    PMG.NodesEditor.editor = new Rete.NodeEditor(
        PMG.codeName + '@' + PMGNodesEditorSchemaVersion,
        document.querySelector('#pmg-nodes-editor')
    )

    PMG.NodesEditor.editor.use(VueRenderPlugin.default)
    PMG.NodesEditor.editor.use(ConnectionPlugin.default)

}

PMG.NodesEditor.initializeEngine = () => {

    PMG.NodesEditor.engine = new Rete.Engine(
        PMG.codeName + '@' + PMGNodesEditorSchemaVersion
    )

}

PMG.NodesEditor.initializeComponents = () => {

    PMG.NodesEditor.components = {

        "绘制盒子": new DrawBoxReteComponent(),
        "绘制棱柱": new DrawPrismReteComponent(),
        "绘制圆柱": new DrawCylinderReteComponent(),
        "绘制圆锥": new DrawConeReteComponent(),
        "绘制棱锥": new DrawPyramidReteComponent(),
        "绘制球体": new DrawSphereReteComponent(),
        "数量": new NumberReteComponent(),
        "加": new AddReteComponent(),
        "减": new SubtractReteComponent(),
        "指向": new PointReteComponent(),
        "向量": new VectorReteComponent(),
        "推拉": new PushPullReteComponent(),
        "移动": new MoveReteComponent(),
        "旋转": new RotateReteComponent(),
        "比例": new ScaleReteComponent(),
        "复制": new CopyReteComponent()

    }

}

PMG.NodesEditor.registerComponents = () => {

    for (var nodeName in PMG.NodesEditor.components) {

        var component = PMG.NodesEditor.components[nodeName]

        PMG.NodesEditor.editor.register(component)
        PMG.NodesEditor.engine.register(component)

    }

}

PMG.NodesEditor.loadToolbarIcons = () => {
    
    document.querySelectorAll('.toolbar .icon').forEach(toolbarIcon => {

        toolbarIcon.src = PMGNodesEditorIcons['nodes'][toolbarIcon.dataset.nodeName]['path']
        toolbarIcon.title = toolbarIcon.dataset.nodeName

    })

}

PMG.NodesEditor.schemaIsExportable = false

PMG.NodesEditor.exportModelSchema = () => {

    if ( PMG.NodesEditor.schemaIsExportable ) {
        sketchup.redrawParametricEntities(JSON.stringify(PMG.NodesEditor.editor.toJSON()))
    }

}

PMG.NodesEditor.resizeEditorView = () => {
    document.querySelector('#pmg-nodes-editor').style.height = window.innerHeight + 'px'
}

PMG.NodesEditor.addEventListeners = () => {

    PMG.NodesEditor.editor.on('process nodecreated noderemoved connectioncreated connectionremoved', () => {
        
        PMG.NodesEditor.engine.abort().then(() => {
            PMG.NodesEditor.engine.process(PMG.NodesEditor.editor.toJSON())
        })
        
    })

    PMG.NodesEditor.editor.on('nodecreated noderemoved nodedragged connectioncreated connectionremoved', () => {
        PMG.NodesEditor.exportModelSchema()
    })

    PMG.NodesEditor.editor.on('nodecreated', (node) => {

        var nodeElement = node.vueContext.$el

        nodeElement.setAttribute('data-node-id', node.id)

        nodeElement.querySelectorAll('input[type="number"]').forEach(nodeNumberInputElement => {

            nodeNumberInputElement.addEventListener('input', _event => {
                PMG.NodesEditor.exportModelSchema()
            })

        })

        nodeElement.querySelectorAll('input[type="text"]').forEach(nodeTextInputElement => {

            nodeTextInputElement.addEventListener('input', _event => {
                PMG.NodesEditor.exportModelSchema()
            })

        })

        nodeElement.querySelectorAll('input[type="checkbox"]').forEach(nodeCheckBoxInputElement => {

            nodeCheckBoxInputElement.addEventListener('change', _event => {
                PMG.NodesEditor.exportModelSchema()
            })

        })

        nodeElement.querySelectorAll('select').forEach(nodeSelectElement => {

            nodeSelectElement.addEventListener('change', _event => {
                PMG.NodesEditor.exportModelSchema()
            })

        })

        var nodeTitleElement = nodeElement.querySelector('.title')

        nodeTitleElement.addEventListener('dblclick', _event => {

            if ( window.confirm(`您是否要删除 "${node.name}" 这个节点？`) ) {
                PMG.NodesEditor.editor.removeNode(node)
            }
            
        })

        var nodeTitleGradient = 'linear-gradient(0deg,hsla(0,0%,100%,.05) 0,hsla(0,0%,100%,.05)' +
            ' 40%,hsla(0,0%,100%,.19)),radial-gradient(70% 40px at center,' +
            PMGNodesEditorIcons['nodes'][node.name]['color'] + ' 0,rgba(0,0,0,0) 60%)'

        nodeTitleElement.style.backgroundImage = nodeTitleGradient

        var nodeTitleIconElement = document.createElement('img')

        nodeTitleIconElement.src = PMGNodesEditorIcons['nodes'][node.name]['path']
        nodeTitleIconElement.className = 'icon'
        nodeTitleElement.appendChild(nodeTitleIconElement)

    })

    PMG.NodesEditor.editor.on('zoom', ({ source }) => {
        return source !== 'dblclick'
    })

    document.querySelectorAll('.toolbar .icon').forEach(toolbarIcon => {

        toolbarIcon.addEventListener('click', event => {

            var component = PMG.NodesEditor.components[event.currentTarget.dataset.nodeName]

            component.createNode().then(node => {
                node.position = [40, 60]
                PMG.NodesEditor.editor.addNode(node)
            })

        })

    })

    window.onresize = PMG.NodesEditor.resizeEditorView

}

PMG.NodesEditor.importModelSchema = () => {

    PMG.NodesEditor.engine.abort().then(() => {

        PMG.NodesEditor.editor.fromJSON(PMGNodesEditorSchema).then(() => {

            PMG.NodesEditor.engine.process(PMG.NodesEditor.editor.toJSON()).then(() => {
                PMG.NodesEditor.schemaIsExportable = true
            })

        })
        
    })

}

PMG.NodesEditor.activateContextMenu = () => {

    var contextMenuOptions = [
        {
            name: '导入 schema 文件',
            fn: () => { sketchup.importSchemaFromFile() }
        },
        {
            name: '导出为 schema 文件',
            fn: () => { sketchup.exportSchemaToFile() }
        },
        {
            name: '冻结参数化实体',
            fn: () => { sketchup.freezeParametricEntities() }
        }
    ]

    new ContextMenu('#pmg-nodes-editor', contextMenuOptions)

}

document.addEventListener('DOMContentLoaded', _event => {

    PMG.NodesEditor.initializeEditor()
    PMG.NodesEditor.initializeEngine()
    PMG.NodesEditor.initializeComponents()
    PMG.NodesEditor.registerComponents()
    PMG.NodesEditor.loadToolbarIcons()
    PMG.NodesEditor.addEventListeners()
    PMG.NodesEditor.importModelSchema()
    PMG.NodesEditor.activateContextMenu()
    PMG.NodesEditor.resizeEditorView()

})
